"""
    Temporal values and comparison.
"""
from __future__ import annotations
import typing
import numpy as np
import numpy.typing as npt

__all__ = [
    "DISTANT_PAST",
    "FAR_FUTURE",
    "NEVER",
    "isnever"
]


@typing.overload
def isnever(t: float) -> bool:
    """
    Returns whether the value of t corresponds to "never". As "never" is implemented as a floating-point NaN,
    direct comparison will always fail, since NaN != NaN.
    """

@typing.overload
def isnever(t: npt.NDArray[np.float64] | list[float]) -> npt.NDArray[np.bool8]:
    """
    Returns an array of booleans corresponding to whether the element of an array correspond to "never". As "never" is
    implemented as a floating-point NaN, direct comparison will always fails, since NaN != NaN.
    """
DISTANT_PAST: float # value = -inf
"""A value that compares less than any other value but itself and NEVER"""

FAR_FUTURE: float # value = inf
"""A value that compares greater than any other value but itself and NEVER"""

NEVER: float # value = nan
"""A value that compares unequal to any value, including itself"""



"""
    Submodule for basic MPI environment discovery.
"""
from __future__ import annotations
from typing import Any

__all__ = [
    "COMM",
    "RANK",
    "SIZE"
]

COMM: Any
"""The MPI communicator if neworder has been installed with the parallel option, otherwise None."""

RANK: int
"""The MPI process rank. 0 in serial mode."""

SIZE: int
"""The number of MPI processes. 1 in serial mode"""



from __future__ import annotations
from typing import Any, Generator

try:
  import networkx as nx  # type: ignore[import]
  import osmnx as ox  # type: ignore[import]
  from shapely.ops import linemerge  # type: ignore[import]
  from shapely.geometry import LineString, MultiLineString, Polygon  # type: ignore[import]
  import geopandas as gpd  # type: ignore[import]
except ImportError:
  raise ImportError("""optional dependencies are not installed.
Reinstalling neworder with the geospatial option should fix this:
pip install neworder[geospatial]""")


class GeospatialGraph:
  """
  Spatial domains on Earth's surface that are defined by graphs/networks.
  Use of this class requires "geospatial" extras: pip install neworder[geospatial]
  """

  def __init__(self, G: nx.Graph, crs: str | None = None) -> None:
    if crs:
      self.__graph = ox.project_graph(G, to_crs=crs)
    else:
      self.__graph = G
    self.__nodes, self.__edges = ox.graph_to_gdfs(self.__graph)

  @classmethod
  def from_point(cls, point: tuple[float, float], *args: Any, crs: str | None = None,  **kwargs: Any) -> GeospatialGraph:
    G = ox.graph_from_point(point, *args, **kwargs)
    return cls(G, crs)

  @property
  def crs(self) -> str:
    return self.__graph.graph["crs"]

  @property
  def graph(self) -> nx.MultiDiGraph | nx.Graph:
    return self.__graph

  @property
  def all_nodes(self) -> gpd.GeoDataFrame:
    return self.__nodes

  @property
  def all_edges(self) -> gpd.GeoDataFrame:
    return self.__edges

  def edges_to(self, node: int) -> Generator[list[tuple[int, int]], None, None]:
    return self.__graph.in_edges(node)

  def edges_from(self, node: int) -> Generator[list[tuple[int, int]], None, None]:
    return self.__graph.out_edges(node)

  def shortest_path(self, origin: int, dest: int, **kwargs: Any) -> LineString:
    nodes = nx.shortest_path(self.__graph, origin, dest, **kwargs)
    route_segments = [self.__edges.loc[(nodes[i], nodes[i+1], 0), "geometry"] for i in range(len(nodes) - 1)]
    return linemerge(MultiLineString(route_segments))

  def subgraph(self, origin: int, **kwargs: Any) -> nx.Graph:
    return nx.ego_graph(self.__graph, origin, **kwargs)

  def isochrone(self, origin: int, **kwargs: Any) -> Polygon:
    subgraph = nx.ego_graph(self.__graph, origin, **kwargs)
    nodes, _ = ox.graph_to_gdfs(subgraph)
    return nodes.geometry.unary_union.convex_hull

